#!/usr/bin/env python2.7
"""
Shared stuff between different modules in this package.  Some
may eventually move to or be replaced by stuff in toil-lib.
"""
from __future__ import print_function
import argparse, sys, os, os.path, random, subprocess, shutil, itertools, glob
import json, timeit, errno
from uuid import uuid4

from toil.common import Toil
from toil.job import Job
from toil.realtimeLogger import RealtimeLogger
from toil_vg.iostore import IOStore
from toil_vg.docker import dockerCall, dockerCheckOutput, _fixPermissions

def add_docker_tool_parse_args(parser):
    """ centralize shared docker options and their defaults """
    parser.add_argument("--no_docker", action="store_true",
                        help="do not use docker for any commands")
    parser.add_argument("--vg_docker", type=str,
                        help="dockerfile to use for vg")
    parser.add_argument("--bcftools_docker", type=str,
                        help="dockerfile to use for bcftools")
    parser.add_argument("--tabix_docker", type=str,
                        help="dockerfile to use for tabix")
    parser.add_argument("--jq_docker", type=str,
                        help="dockerfile to use for jq")

def add_common_vg_parse_args(parser):
    """ centralize some shared io functions and their defaults """
    parser.add_argument('--config', default=None, type=str,
                        help='Config file.  Use toil-vg generate-config to see defaults/create new file')
    
    parser.add_argument("--force_outstore", action="store_true",
                        help="use output store instead of toil for all intermediate files (use only for debugging)")
                        
    parser.add_argument("--chroms", nargs='+',
                        help="Name(s) of reference path in graph(s) (separated by space).  If --graphs "
                        " specified, must be same length/order as --chroms")

    
def get_docker_tool_map(options):
    """ convenience function to parse the above _docker options into a dictionary """

    dmap = dict()
    if not options.no_docker:
        dmap["vg"] = options.vg_docker
        dmap["bcftools"] = options.bcftools_docker
        dmap["tabix"] = options.tabix_docker
        dmap["bgzip"] = options.tabix_docker
        dmap["jq"] = options.jq_docker

    # to do: could be a good place to do an existence check on these tools

    return dmap
        
class DockerRunner(object):
    """ Helper class to centralize docker calling.  So we can toggle Docker
on and off in just one place.  to do: Should go somewhere more central """
    def __init__(self, docker_tool_map = {}):
        # this maps a command to its full docker name
        # example:  docker_tool_map['vg'] = 'quay.io/ucsc_cgl/vg:latest'
        self.docker_tool_map = docker_tool_map

    def has_tool(self, tool):
        # return true if we have an image for this tool
        return tool in self.docker_tool_map

    def call(self, job, args, work_dir = '.' , outfile = None, errfile = None,
             check_output = False, inputs=[]):
        """ run a command.  decide to use docker based on whether
        its in the docker_tool_map.  args is either the usual argument list,
        or a list of lists (in the case of a chain of piped commands)  """
        # from here on, we assume our args is a list of lists
        if len(args) == 0 or len(args) > 0 and type(args[0]) is not list:
            args = [args]
        # convert everything to string
        for i in range(len(args)):
            args[i] = [str(x) for x in args[i]]
        if args[0][0] in self.docker_tool_map:
            return self.call_with_docker(job, args, work_dir, outfile, errfile, check_output, inputs)
        else:
            return self.call_directly(args, work_dir, outfile, errfile, check_output, inputs)
        
    def call_with_docker(self, job, args, work_dir, outfile, errfile, check_output, inputs): 
        """ Thin wrapper for docker_call that will use internal lookup to
        figure out the location of the docker file.  Only exposes docker_call
        parameters used so far.  expect args as list of lists.  if (toplevel)
        list has size > 1, then piping interface used """

        RealtimeLogger.info("Docker Run: {}".format(" | ".join(" ".join(x) for x in args)))
        start_time = timeit.default_timer()

        tool = str(self.docker_tool_map[args[0][0]][0])
        
        if len(args) == 1:
            # one command: we check entry point (stripping first arg if necessary)
            # and pass in single args list
            if len(self.docker_tool_map[args[0][0]]) == 2:
                if self.docker_tool_map[args[0][0]][1]:
                    # not all tools have consistant entrypoints. vg and rtg have entrypoints
                    # but bcftools doesn't. This functionality requires the config file to
                    # operate
                    parameters = args[0][1:]
                else:
                    parameters = args[0]
        else:
            # can leave as is for piped interface which takes list of args lists
            # and doesn't worry about entrypoints since everything goes through bash -c
            parameters = args

        dc = dockerCheckOutput if check_output is True else dockerCall
                    
        ret = dc(job, tool, parameters = parameters,
                 workDir=work_dir, outfile = outfile)
        
        # This isn't running through reliably by itself.  Will assume it's
        # because I took docker.py out of toil, and leave here until we revert to
        # toil's docker
        #
        # Note: It's the tabix docker call in merge_vcf_chunks that's problematic
        #       It complains that the container can't be found, so fixPermissions
        #       doesn't get run afterward.  
        #
        _fixPermissions(tool, work_dir)

        end_time = timeit.default_timer()
        run_time = end_time - start_time
        RealtimeLogger.info("Successfully docker ran {} in {} seconds.".format(
            " | ".join(" ".join(x) for x in args), run_time))

        return ret

    def call_directly(self, args, work_dir, outfile, errfile, check_output, inputs):
        """ Just run the command without docker """

        RealtimeLogger.info("Run: {}".format(" | ".join(" ".join(x) for x in args)))
        start_time = timeit.default_timer()

        # this is all that docker_call does with the inputs parameter:
        for filename in inputs:
            assert(os.path.isfile(os.path.join(work_dir, filename)))

        procs = []
        for i in range(len(args)):
            stdin = procs[i-1].stdout if i > 0 else None
            if i == len(args) - 1 and outfile is not None:
                stdout = outfile
            else:
                stdout = subprocess.PIPE
            procs.append(subprocess.Popen(args[i], stdout=stdout, stderr=errfile,
                                          stdin=stdin, cwd=work_dir))
            
        for p in procs[:-1]:
            p.stdout.close()

        output, errors = procs[-1].communicate()
        for i, proc in enumerate(procs):
            sts = proc.wait()
            if sts != 0:            
                raise Exception("Command {} returned with non-zero exit status {}".format(
                    " ".join(args[i]), sts))

        end_time = timeit.default_timer()
        run_time = end_time - start_time
        RealtimeLogger.info("Successfully ran {} in {} seconds.".format(
            " | ".join(" ".join(x) for x in args), run_time))            

        if check_output:
            return output

def get_files_by_file_size(dirname, reverse=False):
    """ Return list of file paths in directory sorted by file size """

    # Get list of files
    filepaths = []
    for basename in os.listdir(dirname):
        filename = os.path.join(dirname, basename)
        if os.path.isfile(filename):
            filepaths.append(filename)

    # Re-populate list with filename, size tuples
    for i in xrange(len(filepaths)):
        filepaths[i] = (filepaths[i], os.path.getsize(filepaths[i]))

    return filepaths

def clean_toil_path(path):
    """ Try to make input path into something toil friendly """
    # local path
    if ':' not in path:
        return 'file://' + os.path.abspath(path)
    else:
        return path

def import_to_store(toil, options, path, use_out_store = None,
                    out_store_key = None):
    """
    Imports a path into the File or IO store

    Abstract all store writing here so we can switch to the out_store
    when we want to checkpoint an intermeidate file for output
    or just have all intermediate files in the outstore for debugging.
    
    Returns the id in job's file store if use_out_store is True
    otherwise a key (up to caller to make sure its unique)
    in the out_store

    By default options.force_outstore is used to toggle between file and 
    i/o store.  This will be over-ridden by the use_out_store parameter 
    if the latter is not None
    """
    if use_out_store is True or (use_out_store is None and options.force_outstore is True):
        return write_to_store(None, options, path, use_out_store, out_store_key)
    else:
        return toil.importFile(clean_toil_path(path))
    
def write_to_store(job, options, path, use_out_store = None,
                   out_store_key = None):
    """
    Writes path into the File or IO store (from options.out_store)

    Abstract all store writing here so we can switch to the out_store
    when we want to checkpoint an intermeidate file for output
    or just have all intermediate files in the outstore for debugging.
    
    Returns the id in job's file store if use_out_store is True
    otherwise a key (up to caller to make sure its unique)
    in the out_store

    By default options.force_outstore is used to toggle between file and 
    i/o store.  This will be over-ridden by the use_out_store parameter 
    if the latter is not None
    """
    if use_out_store is True or (use_out_store is None and options.force_outstore is True):
        out_store = IOStore.get(options.out_store)
        key = os.path.basename(path) if out_store_key is None else out_store_key
        out_store.write_output_file(path, key)
        return key
    else:
        return job.fileStore.writeGlobalFile(path)

def read_from_store(job, options, id_or_key, path = None, use_out_store = None):
    """
    Reads id (or key) from the File store (or IO store from options.out_store) into path

    Abstract all store reading here so we can switch to the out_store
    when we want to checkpoint an intermeidate file for output
    or just have all intermediate files in the outstore for debugging.

    By default options.force_outstore is used to toggle between file and 
    i/o store.  This will be over-ridden by the use_out_store parameter 
    if the latter is not None
    """
    if use_out_store is True or (use_out_store is None and options.force_outstore is True):
        # we can add this interface if we really want by coming up
        # with unique name here, i guess
        assert path is not None
        out_store = IOStore.get(options.out_store)
        return out_store.read_input_file(id_or_key, path)
    else:
        return job.fileStore.readGlobalFile(id_or_key, path)

def write_dir_to_store(job, options, path, use_out_store = None):
    """
    Need directory interface for rocksdb indexes.  Want to avoid overhead
    of tar-ing up as they may be big.  Write individual files instead, and 
    keep track of the names as well as ids (returns list of name/id pairs)
    """
    out_pairs = []
    file_list = [f for f in os.listdir(path) if os.path.isfile(os.path.join(path, f))]
    for f in file_list:
        f_id = write_to_store(job, options, f, use_out_store = use_out_store,
                              out_store_key = path.replace('/', '_'))
        out_pairs.append(os.path.basename(f), f_id)
    return out_pairs

def read_dir_from_store(job, options, name_id_pairs, path = None, use_out_store = None):
    """
    Need directory interface for rocksdb indexes.  Want to avoid overhead
    of tar-ing up as they may be big.  Takes as input list of filename/id pairs
    and reads these into the local directory given
    """
    if not os.path.isdir(path):
        os.mkdir(path)

    for name, key in name_id_pairs:
        read_from_store(job, options, key, os.path.join(path, name),
                        use_out_store = use_out_store)
    

def batch_iterator(iterator, batch_size):
    """Returns lists of length batch_size.

    This can be used on any iterator, for example to batch up
    SeqRecord objects from Bio.SeqIO.parse(...), or to batch
    Alignment objects from Bio.AlignIO.parse(...), or simply
    lines from a file handle.

    This is a generator function, and it returns lists of the
    entries from the supplied iterator.  Each list will have
    batch_size entries, although the final list may be shorter.
    
    From http://biopython.org/wiki/Split_large_file
    """
    entry = True  # Make sure we loop once
    while entry:
        batch = []
        while len(batch) < batch_size:
            try:
                entry = iterator.next()
            except StopIteration:
                entry = None
            if entry is None:
                # End of file
                break
            batch.append(entry)
        if batch:
            yield batch

def require(expression, message):
    if not expression:
        raise UserError('\n\n' + message + '\n\n')
            



#!/usr/bin/env python2.7
"""
Thin wrapper for vcfeval.  Expects compressed indexed vcfs as input. 
Doesn't really fit in at the moment, but plan on extending into more 
fully fleshed out evaluation phase. 
"""
from __future__ import print_function
import argparse, sys, os, os.path, random, subprocess, shutil, itertools, glob
import json, timeit, errno
from uuid import uuid4

from toil.common import Toil
from toil.job import Job
from toil_vg.vg_common import *

def vcfeval_subparser(parser):
    """
    Create a subparser for vcf evaluation.  Should pass in results of subparsers.add_parser()
    """

    # Add the Toil options so the job store is the first argument
    Job.Runner.addToilOptions(parser)

    # General options
    parser.add_argument("call_vcf", type=str,
                        help="input vcf (must be bgzipped and have .tbi")
    parser.add_argument("truth_vcf", type=str,
                        help="truth vcf (must be bgzipped and have .tbi")
    parser.add_argument("fasta", type=str,
                        help="fasta file containing the DNA sequence.  Can be"
                        " gzipped with .gz extension")
    parser.add_argument("out_store", type=str,
                        help="vcfeval output store")

    # Add common options shared with everybody
    add_common_vg_parse_args(parser)

    # Add docker options
    add_vcfeval_docker_tool_parse_args(parser)

    # Add common calling options shared with vg_evaluation_pipeline
    vcfeval_parse_args(parser)

    # Add common docker options shared with vg_evaluation pipeline
    add_docker_tool_parse_args(parser)

def vcfeval_parse_args(parser):
    """ centralize calling parameters here """

    parser.add_argument("--bed_regions", type=int,
                        help="BED file of regions to consider")
    parser.add_argument("--vcfeval_filtering_opts", type=str,
                        help="Additional filtering options for vcfeval",
                        default="")
    parser.add_argument("--vcfeval_cores", type=int,
                        default=1,
                        help="Cores to use for vcfeval")

def add_vcfeval_docker_tool_parse_args(parser):
    """ vcfeval specific docker options and their defaults """
    parser.add_argument("--rtg_docker", type=list, default=['realtimegenomics/rtg-tools:3.7.1', True],
                        help="dockerfile for rtg_tools")

def get_vcfeval_docker_tool_map(options):
    """ convenience function to parse the above _docker options into a dictionary """

    dmap = dict()
    if not options.no_docker:
        dmap["rtg"] = options.rtg_docker

    return dmap
    
    
def vcfeval(job, work_dir, call_vcf_name, truth_vcf_name,
            sdf_name, outdir_name, bed_name, options):

    """ create and run the vcfeval command """

    cmd = ['rtg', 'vcfeval', '--calls', call_vcf_name,
           '--baseline', truth_vcf_name,
           '--template', sdf_name, '--output', outdir_name,
           '--threads', str(options.vcfeval_cores)]

    if bed_name is not None:
        cmd += ['--evaluation-regions', os.path.basename(bed_name)]

    if len(options.vcfeval_filtering_opts) > 0:
        cmd += options.vcfeval_filtering_opts.split()

    options.drunner.call(job, cmd, work_dir=work_dir)

    # get the F1 out of summary.txt
    # expect header on 1st line and data on 3rd
    # todo: be more robust
    with open(os.path.join(work_dir, os.path.basename(outdir_name), "summary.txt")) as sum_file:
        header = sum_file.readline().split()
        line = sum_file.readline()
        data = sum_file.readline().split()
        assert header[-1] == 'F-measure'
        assert len(data) == len(header)
        return float(data[-1])    
    
def run_vcfeval(job, options,
                truth_vcf_name, truth_vcf_id, truth_tbi_id,
                call_vcf_name, call_vcf_id, call_tbi_id,
                fasta_name, fasta_id,
                bed_name, bed_id):                
    """ run vcf_eval, return f1 score """

    # make a local work directory
    work_dir = job.fileStore.getLocalTempDir()

    # download into local temp directory
    for id,name in [(truth_vcf_id, truth_vcf_name),
                    (truth_tbi_id, truth_vcf_name + '.tbi'),
                    (call_vcf_id, call_vcf_name),
                    (call_tbi_id, call_vcf_name + '.tbi'),
                    (fasta_id, fasta_name),
                    (bed_id, bed_name)]:
        if id is not None:
            job.fileStore.readGlobalFile(id, os.path.join(work_dir, name))

    # output directory
    out_name = "vcfeval_output"
    # indexed sequence
    sdf_name = fasta_name + ".sdf"
    
    # temporary (i hope) hack:  for some reason rtg wants /data/ on the
    # file paths, while most other images don't seem to care.  add them
    # here for now but would really like to understand why
    if options.drunner.has_tool("rtg"):
        fasta_name = os.path.join('/data', fasta_name)
        sdf_name = os.path.join('/data', sdf_name)
        call_vcf_name = os.path.join('/data', call_vcf_name)
        truth_vcf_name = os.path.join('/data', truth_vcf_name)
        out_name = os.path.join('/data', out_name)
        if bed_name is not None:
            bed_name = os.path.join('/data', bed_name)

    # make an indexed sequence (todo: allow user to pass one in)
    options.drunner.call(job, ['rtg', 'format',  fasta_name, '-o', sdf_name], work_dir=work_dir)    

    # run the vcf_eval command
    f1 = vcfeval(job, work_dir, call_vcf_name, truth_vcf_name,
                 sdf_name, out_name, bed_name, options)

    # todo : write everything to output store.
    # (just sticking the f1 score in a text file for now)
    f1_path = os.path.join(work_dir, "f1.txt")
    with open(f1_path, "w") as f:
        f.write(str(f1))    
    out_store = IOStore.get(options.out_store)   
    out_store.write_output_file(f1_path, "f1.txt")

    return f1

def vcfeval_main(options):
    """ command line access to toil vcf eval logic"""
    
    # make the docker runner
    options.drunner = DockerRunner(
        docker_tool_map = dict(get_docker_tool_map(options).items() +
                               get_vcfeval_docker_tool_map(options).items()))
        
    # How long did it take to run the entire pipeline, in seconds?
    run_time_pipeline = None
        
    # Mark when we start the pipeline
    start_time_pipeline = timeit.default_timer()

    # we can relax this down the road by optionally doing some compression/indexing
    assert options.truth_vcf.endswith(".vcf.gz")
    assert options.call_vcf.endswith(".vcf.gz")
    
    with Toil(options) as toil:
        if not toil.options.restart:

            truth_vcf_name = "truth_" + os.path.basename(options.truth_vcf)
            call_vcf_name = "call_" + os.path.basename(options.call_vcf)
            fasta_name = "fa_" + os.path.basename(options.fasta)
            bed_name = "bed_" + os.path.basename(options.bed_regions) if options.bed_regions is not None else None

            # Upload local files to the remote IO Store
            truth_vcf_id = import_to_store(toil, options, options.truth_vcf)
            call_vcf_id = import_to_store(toil, options, options.call_vcf)
            truth_tbi_id = import_to_store(toil, options, options.truth_vcf + '.tbi')
            call_tbi_id = import_to_store(toil, options, options.call_vcf + '.tbi')            
            fasta_id = import_to_store(toil, options, options.fasta)
            bed_id = import_to_store(toil, options, options.bed_regions) if options.bed_regions is not None else None
            
            # Make a root job
            root_job = Job.wrapJobFn(run_vcfeval, options,
                                     truth_vcf_name, truth_vcf_id, truth_tbi_id,
                                     call_vcf_name, call_vcf_id, call_tbi_id,
                                     fasta_name, fasta_id, bed_name, bed_id,
                                     cores=2, memory="5G", disk="2G")
            
            # Run the job
            f1 = toil.start(root_job)
        else:
            f1 = toil.restart()

        print("F1 Score : {}".format(f1))
                
    end_time_pipeline = timeit.default_timer()
    run_time_pipeline = end_time_pipeline - start_time_pipeline
 
    print("All jobs completed successfully. Pipeline took {} seconds.".format(run_time_pipeline))
    
    



import logging
import shlex
import shutil
import subprocess
import tempfile
import textwrap
import filecmp
import vcf
from contextlib import closing
from unittest import TestCase, skip
from urlparse import urlparse
from uuid import uuid4

import os
import posixpath
from bd2k.util.iterables import concat
from boto.s3.connection import S3Connection, Bucket
from boto.s3.key import Key
from toil_vg.iostore import IOStore

log = logging.getLogger(__name__)


class VGCGLTest(TestCase):
    """
    These tests *can* be parameterized with the following optional environment variables:

    TOIL_SCRIPTS_TEST_TOIL_OPTIONS - a space-separated list of additional command line arguments to pass to Toil via
    the script entry point. Default is the empty string.

    TOIL_SCRIPTS_TEST_JOBSTORE - the job store locator to use for the tests. The default is a file: locator pointing
    at a local temporary directory.

    TOIL_SCRIPTS_TEST_NUM_SAMPLES - the number of sample lines to generate in the input manifest
    """

    @classmethod
    def setUpClass(cls):
        super(VGCGLTest, cls).setUpClass()
        # FIXME: pull up into common base class
        logging.basicConfig(level=logging.INFO)

    def _download_input(self, filename, local_path = None):
        tgt = os.path.join(self.workdir, filename) if local_path is None else local_path
        with open(tgt, 'w') as f:
            self.bucket.get_key('/vg_cgl/ci/{}'.format(filename)).get_contents_to_file(f)
        
    def setUp(self):
        self.workdir = tempfile.mkdtemp()
        self.jobStoreAWS = 'aws:us-west-2:testvg-{}'.format(uuid4())
        self.jobStoreLocal = '{}/local-testvg-{}'.format(self.workdir, uuid4())
        self.connection = S3Connection()
        self.bucket = self.connection.get_bucket('cgl-pipeline-inputs')
                
        self.base_command = concat('toil-vg', 'run',
                                   '--realTimeLogging', '--logInfo', '--reads_per_chunk', '8000',
                                   '--call_chunk_size', '20000',
                                   '--index_mode', 'gcsa-mem', '--index_cores', '7', '--alignment_cores', '4',
                                   '--calling_cores', '4')
        
        # default output store
        self.outstore = 'aws:us-west-2:toilvg-jenkinstest-outstore-{}'.format(uuid4())
        self.local_outstore = os.path.join(self.workdir, 'toilvg-jenkinstest-outstore-{}'.format(uuid4()))
                
        # copy the sequence information for vcf comparison
        # (lumped in one file out of laziness.  todo: at least split by chromosome)
        self.chrom_fa = os.path.join(self.workdir, 'chrom.fa.gz')
        self._download_input('chrom.fa.gz', self.chrom_fa)

        
    def test_chr17_sampleNA12877(self):
        ''' Test sample BRCA1 output, graph construction and use, and local file processing
        '''
        self._download_input('NA12877.brca1.bam.fq.gz')
        self._download_input('BRCA1_chrom_name_chop_100.vg')
        
        self.sample_reads = os.path.join(self.workdir, 'NA12877.brca1.bam.fq.gz')
        self.test_vg_graph = os.path.join(self.workdir, 'BRCA1_chrom_name_chop_100.vg')
        self._run(self.base_command, self.jobStoreLocal, self.sample_reads, 'NA12877',
                  self.local_outstore, '--graphs',  self.test_vg_graph, '--chroms', '17',
                  '--call_opts', '--offset 43044293', '--interleaved')

        self._assertOutput('NA12877_17.vcf', self.local_outstore)
    

    def test_chr19_sampleNA12877(self):
        ''' Test sample LRC KIR output
        '''
        
        self._download_input('NA12877.lrc_kir.bam.small.fq')
        self._download_input('LRC_KIR_chrom_name_chop_100.small.vg')
        self._download_input('LRC_KIR_chrom_name_chop_100.small.vg.xg')
        self._download_input('LRC_KIR_chrom_name_chop_100.small.vg.gcsa')
        self._download_input('LRC_KIR_chrom_name_chop_100.small.vg.gcsa.lcp')
        self.sample_reads = os.path.join(self.workdir, 'NA12877.lrc_kir.bam.small.fq')
        self.test_vg_graph = os.path.join(self.workdir, 'LRC_KIR_chrom_name_chop_100.small.vg')
        self.test_xg_index = os.path.join(self.workdir, 'LRC_KIR_chrom_name_chop_100.small.vg.xg')
        self.test_gcsa_index = os.path.join(self.workdir, 'LRC_KIR_chrom_name_chop_100.small.vg.gcsa')
        
        self._run(self.base_command, self.jobStoreLocal, self.sample_reads, 'NA12877',
                  self.local_outstore,  '--gcsa_index', self.test_gcsa_index,
                  '--xg_index', self.test_xg_index, '--graphs', self.test_vg_graph,
                  '--chroms', '19', '--force_outstore')
        self._assertOutput('NA12877_19.vcf', self.local_outstore)

    def test_chr6_MHC_sampleNA12877(self):
        ''' Test sample MHC output
        '''
        self.sample_reads = 's3://cgl-pipeline-inputs/vg_cgl/ci/NA12877.mhc.bam.small.fq'
        self.test_vg_graph = 's3://cgl-pipeline-inputs/vg_cgl/ci/MHC_chrom_name_chop_100.small.vg'
        self.test_gcsa_index = 's3://cgl-pipeline-inputs/vg_cgl/ci/MHC_chrom_name_chop_100.small.vg.gcsa'
        self._run(self.base_command, self.jobStoreLocal, self.sample_reads, 'NA12877',
                  self.local_outstore,  '--gcsa_index', self.test_gcsa_index,
                  '--graphs', self.test_vg_graph, '--chroms', '6')
        self._assertOutput('NA12877_6.vcf', self.local_outstore)

    def test_chr5_SMA_sampleNA12877(self):
        ''' Test sample SMA output
        '''
        self._download_input('NA12877.sma.bam.small.fq')
        self._download_input('SMA_chrom_name_chop_100.small.vg')
        self.sample_reads = os.path.join(self.workdir, 'NA12877.sma.bam.small.fq')
        self.test_vg_graph = os.path.join(self.workdir, 'SMA_chrom_name_chop_100.small.vg')
        self._run(self.base_command, self.jobStoreLocal, self.sample_reads, 'NA12877',
                  self.local_outstore, '--graphs', self.test_vg_graph, '--chroms', '5')
        # disabling this test for now as no variants get called,
        # which triggers assert fail when loading the vceval output
        #self._assertOutput('NA12877_5.vcf', self.local_outstore)
    
    def _run(self, *args):
        args = list(concat(*args))
        log.info('Running %r', args)
        subprocess.check_call(args)

    def _assertOutput(self, testFile, outstore, f1_threshold=0.90):
        # todo: clean this to be less hardcoded
        testName = 'NA12877.vcf.gz'
        normalName = 'normal_' + testFile + '.gz'
        localTest = os.path.join(self.workdir, testName)
        localNormal = os.path.join(self.workdir, normalName)

        # copy baseline
        self._download_input(normalName, localNormal)
        self._download_input(normalName + '.tbi', localNormal + '.tbi')

        # copy calls
        io_store = IOStore.get(outstore)
        io_store.read_input_file(testName, localTest)
        io_store.read_input_file(testName + '.tbi', localTest + '.tbi')

        # run vcfeval
        eval_results = os.path.join(self.workdir, 'eval_results')
        subprocess.check_call(['toil-vg', 'vcfeval', self.jobStoreLocal, localTest,
                               localNormal, self.chrom_fa, eval_results,
                               '--vcfeval_cores', '8'])
        with open(os.path.join(eval_results, 'f1.txt')) as f1_file:
            f1_score = float(f1_file.readline().strip())
        
        self.assertTrue(f1_score >= f1_threshold)

    def tearDown(self):
        shutil.rmtree(self.workdir)
        subprocess.check_call(['toil', 'clean', self.jobStoreAWS])
        subprocess.check_call(['toil', 'clean', self.jobStoreLocal])
        



#!/usr/bin/env python2.7
"""
vg_config.py: Default configuration values all here (and only here), as well as logic
for reading and generating config files.

"""
from __future__ import print_function
import argparse, sys, os, os.path, errno, random, subprocess, shutil, itertools, glob, tarfile
import doctest, re, json, collections, time, timeit
import logging, logging.handlers, SocketServer, struct, socket, threading
import string
import urlparse
import getpass
import pdb
import textwrap
import yaml
from toil_vg.vg_common import require

# TODO: configure RPATH-equivalent on OS X for finding libraries without environment variables at runtime
def generate_config():
    return textwrap.dedent("""
        # Toil VG Pipeline configuration file (created by toil-vg generate-config)
        # This configuration file is formatted in YAML. Simply write the value (at least one space) after the colon.
        # Edit the values in the configuration file and then rerun the pipeline: "toil-vg run"
        # 
        # URLs can take the form: "/", "s3://"
        # Local inputs follow the URL convention: "/full/path/to/input.txt"
        # S3 URLs follow the convention: "s3://bucket/directory/file.txt"
        #
        # Comments (beginning with #) do not need to be removed. 
        # Command-line options take priority over parameters in this file.  
        ######################################################################################################################

        ###########################################
        ### Toil resource tuning                ###
        
        # These parameters must be adjusted based on data and cluster size
        # when running on anything other than single-machine mode

        # TODO: Reduce number of parameters here.  Seems fine grained, especially for disk/mem
        # option to spin off config files for small/medium/large datasets?   

        # The following parameters assign resources to small helper jobs that typically don't do 
        # do any computing outside of toil overhead.  Generally do not need to be changed. 
        misc-cores: 1
        misc-mem: '1G'
        misc-disk: '1G'
    
        # Resources allotted for gcsa pruning.  Note that the vg mod commands used in
        # this stage generally cannot take advantage of more than one thread
        prune-cores: 1
        prune-mem: '4G'
        prune-disk: '2G'

        # Resources allotted for xg and gcsa indexing and kmers computing
        index-cores: 1
        index-mem: '4G'
        index-disk: '2G'

        # Resources for fastq splitting and gam merging
        # fastq spilt can use up to 2 cores, gam merging single threaded
        fq-split-cores: 1
        fq-split-mem: '4G'
        fq-split-disk: '2G'

        # Number of threads to use for Rocksdb GAM indexing
        # Generally, this should be kept low as speedup drops off radically 
        # after a few threads.
        gam-index-cores: 1

        # Resources for *each* vg map job
        # the number of vg map jobs is controlled by reads-per-chunk (below)
        alignment-cores: 1
        alignment-mem: '4G'
        alignment-disk: '2G'

        # Resources for chunking up a graph/gam for calling (and merging)
        # typically take xg for whoe grpah, and gam for a chromosome,
        # and split up into chunks of call-chunk-size (below)
        call-chunk-cores: 1
        call-chunk-mem: '4G'
        call-chunk-disk: '2G'

        # Resources for calling each chunk (currently includes pileup/call/genotype)
        calling-cores: 1
        calling-mem: '4G'
        calling-disk: '2G'

        ###########################################
        ### Arguments Shared Between Components ###
        # Use output store instead of toil for all intermediate files (use only for debugging)
        force-outstore: False
        
        #############################
        ### Docker Tool Arguments ###
        # Do not use docker for any commands
        no-docker: False
        
        ## Docker Tool List ##
        ##   Each tool is specified as a list where the first element is the docker image URL,
        ##   and the second element indicates if the docker image has an entrypoint or not
        ##   If left blank or commented, then the tool will be run directly from the command line instead
        ##   of through docker. no-docker (above) overrides all these options. 
        # Dockerfile to use for vg

        vg-docker: ['quay.io/glennhickey/vg:v1.4.0-2134-g161bec5', True]

        # Dockerfile to use for bcftools
        bcftools-docker: ['quay.io/cmarkello/bcftools', False]

        # Dockerfile to use for tabix
        tabix-docker: ['quay.io/cmarkello/htslib:latest', False]

        # Dockerfile to use for jq
        jq-docker: ['devorbitus/ubuntu-bash-jq-curl', False]
        
        # Dockerfile to use for rtg
        rtg-docker: ['realtimegenomics/rtg-tools:3.7.1', True]
        
        ##########################
        ### vg_index Arguments ###

        # Name of index output files.  ex <name>.xg, <name>.gcsa etc. 
        index-name: 'genome'

        # Options to pass to vg mod for pruning phase. (if empty list, phase skipped)
        # The primary path will always be added back onto the pruned grpah
        prune-opts: ['-p', '-l', '16', '-S', '-e', '5']

        # Options to pass to 2nd vg mod for pruning phase.  The input will
        # will be piped in from the prune-opts command above.  Will be 
        # skipped if empty list
        prune-opts-2: ['-S', '-l', '32']

        # Options to pass to vg kmers.
        kmers-opts: ['-g', '-B', '-k', '16', '-H', '1000000000', '-T', '1000000001']

        # Options to pass to vg gcsa indexing
        gcsa-opts: ['-X', '3', '-Z', '3000']

        ########################
        ### vg_map Arguments ###

        # Number of reads per chunk to use when splitting up fastq.  
        # Each chunk will correspond to a vg map job
        reads-per-chunk: 10000000
        
        # Context expansion used for gam chunking
        chunk_context: 50
    
        # Treat input fastq as paired-end interleaved
        interleaved: False
        
        # Core arguments for vg mapping (do not include file names or -t/--threads)
        # Note -i/--interleaved will be ignored. use the --interleaved option 
        # on the toil-vg command line instead
        map-opts: ['-M2', '-W', '500', '-u', '0', '-U', '-O', '-S', '50', '-a']
        
        # Type of vg index to use for mapping (either 'gcsa-kmer' or 'gcsa-mem')
        index-mode: gcsa-mem

        #########################
        ### vg_call Arguments ###
        # Overlap option that is passed into make_chunks and call_chunk
        overlap: 2000
        
        # Chunk size
        call-chunk-size: 10000000

        # Options to pass to chunk_gam. (do not include file names or -t/--threads)
        filter-opts: ['-r', '0.9', '-fu', '-s', '1000', '-o', '0', '-q', '15']

        # Options to pass to vg pileup. (do not include file names or -t/--threads)
        pileup-opts: ['-q', '10', '-a']

        # Options to pass to vg call. (do not include file names or -t/--threads)
        call-opts: ['']
        
        # Options to pass to vg genotype. (do not include file names or -t/--threads)
        genotype-opts: ['']

        # Use vg genotype instead of vg call
        genotype: False
        
    """)


def apply_config_file_args(args):
    """
    Merge args from the config file and the parser, giving priority to the parser.
    """

    # turn --*_opts from strings to lists to be consistent with config file
    for x_opts in ['map_opts', 'call_opts', 'filter_opts', 'genotype_opts']:
        if x_opts in args.__dict__.keys() and type(args.__dict__[x_opts]) is str:
            args.__dict__[x_opts] = args.__dict__[x_opts].split(' ')
            # get rid of any -t or --threads while we're at it
            for t in ['-t', '--threads']:
                if t in args.__dict__[x_opts]:
                    pos = args.__dict__[x_opts].index(t)
                    del args.__dict__[x_opts][pos:pos+2]

    # If no config file given, we generate a default one
    if args.config is None:
        config = generate_config()
    else:
        require(os.path.exists(args.config), 'Config, {}, not found. Please run '
            '"toil-vg generate-config > {}" to create.'.format(args.config, args.config))    
        with open(args.config) as conf:
            config = conf.read()
                
    # Parse config
    parsed_config = {x.replace('-', '_'): y for x, y in yaml.load(config).iteritems()}
    options = argparse.Namespace(**parsed_config)

    # Add in options from the program arguments to the arguments in the config file
    #   program arguments that are also present in the config file will overwrite the
    #   arguments in the config file
    for args_key in args.__dict__:
        # Add in missing program arguments to config option list and
        # overwrite config options with corresponding options that are not None in program arguments
        if (args.__dict__[args_key]) or (args_key not in  options.__dict__.keys()):
            options.__dict__[args_key] = args.__dict__[args_key]
            
    return options

def config_main(args):
    """ config just prints out a file """
    
    sys.stdout.write(generate_config())



#!/usr/bin/env python2.7
"""
vg_index.py: index a graph so it can be mapped to

"""
from __future__ import print_function
import argparse, sys, os, os.path, errno, random, subprocess, shutil, itertools, glob, tarfile
import doctest, re, json, collections, time, timeit
import logging, logging.handlers, SocketServer, struct, socket, threading
import string
import urlparse
import getpass
import pdb
import gzip

from math import ceil
from subprocess import Popen, PIPE

from toil.common import Toil
from toil.job import Job
from toil.realtimeLogger import RealtimeLogger
from toil_vg.vg_common import *

def map_subparser(parser):
    """
    Create a subparser for mapping.  Should pass in results of subparsers.add_parser()
    """

    # Add the Toil options so the job store is the first argument
    Job.Runner.addToilOptions(parser)
    
    # General options
    
    parser.add_argument("sample_reads", type=str,
        help="Path to sample reads in fastq format (.fq.gz also supported)")
    parser.add_argument("sample_name", type=str,
        help="sample name (ex NA12878)")
    parser.add_argument("xg_index", type=str,
        help="Path to xg index")    
    parser.add_argument("gcsa_index", type=str,
        help="Path to GCSA index")
    parser.add_argument("out_store",
        help="output IOStore to create and fill with files that will be downloaded to the local machine where this toil script was run")
    parser.add_argument("--kmer_size", type=int,
        help="size of kmers to use in indexing and mapping")

    # Add common options shared with everybody
    add_common_vg_parse_args(parser)

    # Add mapping options
    map_parse_args(parser)

    # Add common docker options
    add_docker_tool_parse_args(parser)


def map_parse_args(parser, stand_alone = False):
    """ centralize indexing parameters here """

    parser.add_argument("--reads_per_chunk", type=int,
        help="number of reads for each mapping job")
    parser.add_argument("--alignment_cores", type=int,
        help="number of threads during the alignment step")
    parser.add_argument("--index_mode", choices=["gcsa-kmer",
        "gcsa-mem"],
        help="type of vg index to use for mapping")
    parser.add_argument("--interleaved", action="store_true", default=False,
                        help="treat fastq as interleaved read pairs.  overrides map-args")
    parser.add_argument("--map-opts", type=str,
                        help="arguments for vg map (wrapped in \"\")")

def run_mapping(job, options, xg_file_id, gcsa_and_lcp_ids, reads_file_id):
    """ split the fastq, then align each chunk """
    fastq_chunk_ids = job.addChildJobFn(run_split_fastq, options, reads_file_id, cores=options.fq_split_cores,
                                        memory=options.fq_split_mem, disk=options.fq_split_disk).rv()

    return job.addFollowOnJobFn(run_whole_alignment, options, xg_file_id, gcsa_and_lcp_ids,
                                fastq_chunk_ids, cores=options.misc_cores,
                                memory=options.misc_mem, disk=options.misc_disk).rv()    

def run_split_fastq(job, options, sample_fastq_id):
    
    RealtimeLogger.info("Starting fastq split and alignment...")
    start_time = timeit.default_timer()
    
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()

    # We need the sample fastq for alignment
    sample_filename = os.path.basename(options.sample_reads)
    fastq_path = os.path.join(work_dir, os.path.basename(options.sample_reads))
    fastq_gzipped = os.path.splitext(fastq_path)[1] == '.gz'
    read_from_store(job, options, sample_fastq_id, fastq_path)

    # Split up the fastq into chunks

    # Make sure chunk size even in case paired interleaved
    chunk_size = options.reads_per_chunk
    if chunk_size % 2 != 0:
        chunk_size += 1

    # 4 lines per read
    chunk_lines = chunk_size * 4

    # Note we do this on the command line because Python is too slow
    # Assume: gzip installed (can't do this command in docker with current interface)
    uc_fastq_name = fastq_path if not fastq_gzipped else 'input_reads.fq'
    if fastq_gzipped:
        cmd = [['gzip', '-d', '-c', os.path.basename(fastq_path)]]
    else:
        cmd = [['cat', os.path.basename(fastq_path)]]
        
    cmd.append(['split', '-l', str(chunk_lines), '--filter=gzip > $FILE.fq.gz',
                '-', 'fq_chunk.'])

    options.drunner.call_directly(cmd, work_dir, None, None, False, [])

    fastq_chunk_ids = []
    for chunk_name in os.listdir(work_dir):
        if chunk_name.endswith('.fq.gz') and chunk_name.startswith('fq_chunk'):
            fastq_chunk_ids.append(write_to_store(job, options, os.path.join(work_dir, chunk_name)))
        
    end_time = timeit.default_timer()
    run_time = end_time - start_time
    RealtimeLogger.info("Split fastq into {} chunks. Process took {} seconds.".format(len(fastq_chunk_ids), run_time))

    return fastq_chunk_ids

def run_whole_alignment(job, options, xg_file_id, gcsa_and_lcp_ids, fastq_chunk_ids):
    """ align all fastq chunks in parallel
    """
    
    # this will be a list of lists.
    # gam_chunk_file_ids[i][j], will correspond to the jth path (from options.chroms)
    # for the ith gam chunk (generated from fastq shard i)
    gam_chunk_file_ids = []

    for chunk_id, chunk_filename_id in enumerate(fastq_chunk_ids):        
        #Run graph alignment on each fastq chunk
        gam_chunk_ids = job.addChildJobFn(run_chunk_alignment, options, chunk_filename_id, chunk_id,
                                          xg_file_id, gcsa_and_lcp_ids,
                                          cores=options.alignment_cores, memory=options.alignment_mem,
                                          disk=options.alignment_disk).rv()
        gam_chunk_file_ids.append(gam_chunk_ids)

    return job.addFollowOnJobFn(run_merge_gams, options, gam_chunk_file_ids, cores=options.misc_cores,
                                memory=options.misc_mem, disk=options.misc_disk).rv()


def run_chunk_alignment(job, options, chunk_filename_id, chunk_id, xg_file_id, gcsa_and_lcp_ids):

    RealtimeLogger.info("Starting alignment on {} chunk {}".format(options.sample_name, chunk_id))
    # Set up the IO stores each time, since we can't unpickle them on Azure for
    # some reason.
    out_store = IOStore.get(options.out_store)

    # How long did the alignment take to run, in seconds?
    run_time = None
    
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()

    # Download local input files from the remote storage container
    graph_file = os.path.join(work_dir, "graph.vg")

    xg_file = graph_file + ".xg"
    read_from_store(job, options, xg_file_id, xg_file)
    gcsa_file = graph_file + ".gcsa"
    gcsa_file_id = gcsa_and_lcp_ids[0]
    read_from_store(job, options, gcsa_file_id, gcsa_file)
    lcp_file = gcsa_file + ".lcp"
    lcp_file_id = gcsa_and_lcp_ids[1]
    read_from_store(job, options, lcp_file_id, lcp_file)

    # We need the sample fastq for alignment
    fastq_file = os.path.join(work_dir, 'chunk_{}.fq'.format(chunk_id))
    read_from_store(job, options, chunk_filename_id, fastq_file)
    
    # And a temp file for our aligner output
    output_file = os.path.join(work_dir, "{}_{}.gam".format(options.sample_name, chunk_id))

    # Open the file stream for writing
    with open(output_file, "w") as alignment_file:

        # Start the aligner and have it write to the file

        # Plan out what to run
        vg_parts = []
        vg_parts += ['vg', 'map', '-f', os.path.basename(fastq_file), os.path.basename(graph_file),
                     '-t', str(options.alignment_cores)]
        vg_parts += options.map_opts

        # Override the -i flag in args with the --interleaved command-line flag
        if options.interleaved is True and '-i' not in vg_parts and '--interleaved' not in vg_parts:
            vg_parts += ['-i']
        elif options.interleaved is False and 'i' in vg_parts:
            del vg_parts[vg_parts.index('-i')]
        if options.interleaved is False and '--interleaved' in vg_parts:
            del vg_parts[vg_parts.index('--interleaved')]

        if options.index_mode == "gcsa-kmer":
            # Use the new default context size in this case
            vg_parts += ['-x', os.path.basename(xg_file), '-g', os.path.basename(gcsa_file),
                '-n5', '-k', str(options.kmer_size)]
        elif options.index_mode == "gcsa-mem":
            # Don't pass the kmer size, so MEM matching is used
            vg_parts += ['-x', os.path.basename(xg_file), '-g', os.path.basename(gcsa_file), '-n5']
        else:
            raise RuntimeError("invalid indexing mode: " + options.index_mode)

        RealtimeLogger.info(
            "Running VG for {} against {}: {}".format(options.sample_name, graph_file,
            " ".join(vg_parts)))
        
        # Mark when we start the alignment
        start_time = timeit.default_timer()
        command = vg_parts
        options.drunner.call(job, command, work_dir = work_dir, outfile=alignment_file)
        
        # Mark when it's done
        end_time = timeit.default_timer()
        run_time = end_time - start_time

    RealtimeLogger.info("Aligned {}. Process took {} seconds.".format(output_file, run_time))

    # Chunk the gam up by chromosome
    gam_chunks = split_gam_into_chroms(job, work_dir, options, xg_file, output_file)

    # Write gam_chunks to store
    gam_chunk_ids = []
    for gam_chunk in gam_chunks:
        gam_chunk_ids.append(write_to_store(job, options, gam_chunk))

    return gam_chunk_ids

def split_gam_into_chroms(job, work_dir, options, xg_file, gam_file):
    """
    Create a Rocksdb index then use it to split up the given gam file
    (a local path) into a separate gam for each chromosome.  
    Return a list of filenames.  the ith filename will correspond
    to the ith path in the options.path_name list
    """

    output_index = gam_file + '.index'
    
    # Index the alignment by node
    start_time = timeit.default_timer()

    index_cmd = ['vg', 'index', '-N', os.path.basename(gam_file),
                 '-d', os.path.basename(output_index), '-t', str(options.gam_index_cores)]
    options.drunner.call(job, index_cmd, work_dir = work_dir)
        
    end_time = timeit.default_timer()
    run_time = end_time - start_time
    RealtimeLogger.info("Indexed {}. Process took {} seconds.".format(output_index, run_time))

    # Chunk the alignment into chromosomes using the input paths

    # First, we make a list of paths in a format that we can pass to vg chunk
    path_list_name = os.path.join(work_dir, 'path_list')
    with open(path_list_name, 'w') as path_list:
        for path in options.chroms:
            path_list.write('{}\n'.format(path))

    output_bed_name = os.path.join(work_dir, 'output_bed.bed')
    
    # Now run vg chunk on the gam index to get our gams
    chunk_cmd = ['vg', 'chunk', '-x', os.path.basename(xg_file),
                 '-a', os.path.basename(output_index), '-c', str(options.chunk_context),
                 '-P', os.path.basename(path_list_name),
                 '-b', os.path.splitext(os.path.basename(gam_file))[0],
                 '-t', str(options.alignment_cores),
                 '-R', os.path.basename(output_bed_name)]
    
    start_time = timeit.default_timer()

    options.drunner.call(job, chunk_cmd, work_dir = work_dir)
    
    end_time = timeit.default_timer()
    run_time = end_time - start_time
    RealtimeLogger.info("Chunked {}. Process took {} seconds.".format(gam_file, run_time))

    # scrape up the vg chunk results into a list of paths to the output gam
    # chunks and return them.  we expect the filenames to be in the 4th column
    # of the output bed from vg chunk. 
    gam_chunks = []
    with open(output_bed_name) as output_bed:
        for line in output_bed:
            toks = line.split('\t')
            if len(toks) > 3:
                gam_chunks.append(os.path.join(work_dir, os.path.basename(toks[3].strip())))
                assert os.path.splitext(gam_chunks[-1])[1] == '.gam'
    assert len(gam_chunks) == len(options.chroms)
    return gam_chunks


def run_merge_gams(job, options, gam_chunk_file_ids):
    """
    Merge together gams, doing each chromosome in parallel
    """

    chr_gam_ids = []

    for i, chr in enumerate(options.chroms):
        shard_ids = [gam_chunk_file_ids[j][i] for j in range(len(gam_chunk_file_ids))]
        chr_gam_id = job.addChildJobFn(run_merge_chrom_gam, options, chr, shard_ids,
                                       cores=options.fq_split_cores,
                                       memory=options.fq_split_mem,
                                       disk=options.fq_split_disk).rv()
        chr_gam_ids.append(chr_gam_id)
    
    return chr_gam_ids


def run_merge_chrom_gam(job, options, chr_name, chunk_file_ids):
    """
    Make a chromosome gam by merging up a bunch of gam ids, one 
    for each  shard.  
    """
    # Define work directory for docker calls
    work_dir = job.fileStore.getLocalTempDir()
    
    output_file = os.path.join(work_dir, '{}_{}.gam'.format(options.sample_name, chr_name))

    # Would be nice to be able to do this merge with fewer copies.. 
    with open(output_file, 'a') as merge_file:
        for chunk_gam_id in chunk_file_ids:
            tmp_gam_file = os.path.join(work_dir, 'tmp_{}.gam'.format(uuid4()))
            read_from_store(job, options, chunk_gam_id, tmp_gam_file)
            with open(tmp_gam_file) as tmp_f:
                shutil.copyfileobj(tmp_f, merge_file)
                
    chr_gam_id = write_to_store(job, options, output_file)

    # checkpoint to out store
    if not options.force_outstore or options.tool == 'map':
        write_to_store(job, options, output_file, use_out_store = True)
            
    return chr_gam_id

def map_main(options):
    """
    Wrapper for vg map. 
    """

    # make the docker runner
    options.drunner = DockerRunner(
        docker_tool_map = get_docker_tool_map(options))

    # Some file io is dependent on knowing if we're in the pipeline
    # or standalone. Hack this in here for now
    options.tool = 'map'

    # Throw error if something wrong with IOStore string
    IOStore.get(options.out_store)
    
    # How long did it take to run the entire pipeline, in seconds?
    run_time_pipeline = None
        
    # Mark when we start the pipeline
    start_time_pipeline = timeit.default_timer()
    
    with Toil(options) as toil:
        if not toil.options.restart:
            
            # Upload local files to the remote IO Store
            inputXGFileID = import_to_store(toil, options, options.xg_index)
            inputGCSAFileID = import_to_store(toil, options, options.gcsa_index)
            inputLCPFileID = import_to_store(toil, options, options.gcsa_index + ".lcp")
            sampleFastqFileID = import_to_store(toil, options, options.sample_reads)
            
            # Make a root job
            root_job = Job.wrapJobFn(run_mapping, options,inputXGFileID,
                                     (inputGCSAFileID, inputLCPFileID), sampleFastqFileID,
                                     cores=options.misc_cores,
                                     memory=options.misc_mem,
                                     disk=options.misc_disk)
            
            # Run the job and store the returned list of output files to download
            toil.start(root_job)
        else:
            toil.restart()
            
    end_time_pipeline = timeit.default_timer()
    run_time_pipeline = end_time_pipeline - start_time_pipeline
 
    print("All jobs completed successfully. Pipeline took {} seconds.".format(run_time_pipeline))
    
